import json

from django.core.management.base import BaseCommand

from emeis.core.models import ACL, Role, Scope, User
from emeis.core.utils import rebuild_mptt_model


class Command(BaseCommand):
    """Create an Emeis scope."""

    help = "Create an Emeis scope."

    def add_arguments(self, parser):
        parser.add_argument(
            "--parent",
            "-p",
            dest="parent",
            type=str,
            required=False,
            help="ID of parent scope. If not given, the new scope will be a top-level entity",
        )
        parser.add_argument(
            "--user",
            "-u",
            dest="user",
            type=str,
            required=False,
            help="Username to assign to the scope. User must exist. Only works with --role",
        )
        parser.add_argument(
            "--role",
            "-r",
            dest="role",
            type=str,
            required=False,
            help="Role to assign user. Only works with --user",
        )
        parser.add_argument(
            "--name",
            "-n",
            dest="name",
            type=str,
            required=True,
            help="Name of the new scope. Can be either a text, or a JSON object for a multilingual name",
        )
        parser.add_argument(
            "--json-out",
            dest="json_out",
            required=False,
            default=False,
            action="store_true",
            help="If given, output will be JSON instead of text. Useful for scripting",
        )

    def handle(self, *args, **options):
        out_info = {}
        try:
            parent = (
                Scope.objects.get(pk=options["parent"]) if options["parent"] else None
            )
            if parent:
                out_info["parent_scope_id"] = str(parent.pk)
        except Scope.DoesNotExist:
            self.stderr.write(
                f"Parent scope with given ID not found: {options['parent']}"
            )
            return

        if options["user"] and not options["role"]:
            self.stderr.write("If you pass --user, you must also pass --role")
            return
        elif options["role"] and not options["user"]:
            self.stderr.write("If you pass --role, you must also pass --user")
            return

        if options["name"].startswith("{"):
            options["name"] = json.loads(options["name"])

        new_scope = Scope.objects.create(name=options["name"], parent=parent)
        rebuild_mptt_model(Scope)
        out_info["scope_id"] = str(new_scope.pk)

        if options["user"]:
            user = User.objects.get(username=options["user"])
            role = Role.objects.get(pk=options["role"])
            new_acl = ACL.objects.create(user=user, role=role, scope=new_scope)
            out_info["acl_id"] = str(new_acl.pk)

        if options["json_out"]:
            self.stdout.write(json.dumps(out_info, indent=4))
        else:
            # string output
            if options["user"]:
                self.stdout.write(
                    f"ACL for new scope created: user={options['user']}, role={role.pk}"
                )
            self.stdout.write(f"Scope created: id={new_scope.pk}")
        self.stdout.write("\n")



import pytest
from django.urls import reverse
from rest_framework import status

from emeis.core.models import User


def test_search_users(
    admin_client,
    acl_factory,
):

    users_list = [acl.user for acl in acl_factory.create_batch(5)]

    resp = admin_client.get(
        reverse("user-list"), {"filter[search]": users_list[0].first_name}
    )
    returned_user_ids = [us["id"] for us in resp.json()["data"]]

    assert str(users_list[0].pk) in returned_user_ids


@pytest.mark.parametrize(
    "filter_name, expect_result",
    [("hasRole", True), ("has_role", True), ("hasrole", False)],
)
def test_user_has_role(admin_client, acl_factory, filter_name, expect_result):
    users_list = [acl.user for acl in acl_factory.create_batch(3)]

    resp = admin_client.get(
        reverse("user-list"),
        {f"filter[{filter_name}]": users_list[2].acls.first().role_id},
    )

    if expect_result:
        ret_users = [us["attributes"]["username"] for us in resp.json()["data"]]
        expected = [str(users_list[2].username)]

        assert expected == ret_users
    else:
        assert resp.status_code == status.HTTP_400_BAD_REQUEST


@pytest.mark.parametrize(
    "filter_field, model_attr, expect_result",
    [
        ("id__in", "pk", True),
        ("id", "pk", True),
        ("username", "username", True),
        ("username", "email", False),
        ("email__in", "email", True),
        ("email__in", "username", False),
    ],
)
def test_declared_filters(
    admin_client, user_factory, filter_field, model_attr, expect_result
):
    user1, _ = user_factory.create_batch(2)

    attr_value = getattr(user1, model_attr)

    resp = admin_client.get(
        reverse("user-list"), {f"filter[{filter_field}]": attr_value}
    )

    ret_users = [us["attributes"]["username"] for us in resp.json()["data"]]

    if expect_result:
        assert [user1.username] == ret_users
    else:
        assert ret_users == []


@pytest.mark.parametrize("sort", ["email", "-email"])
def test_user_ordering_case_insensitive(admin_client, admin_user, user_factory, sort):
    emails = [
        "Aaaa@example.com",
        "Zzzzz@example.com",
        "aaaaa@example.com",
        "m@example.com",
    ]
    for email in emails:
        user_factory.create(email=email)

    resp = admin_client.get(reverse("user-list"), {"sort": sort})

    expect_emails = sorted(
        emails + [admin_user.user.email],
        key=lambda e: e.lower(),
        reverse=sort.startswith("-"),
    )

    assert expect_emails == [d["attributes"]["email"] for d in resp.json()["data"]]


@pytest.mark.parametrize("sort", ["username", "-username"])
def test_user_ordering_case_sensitive(admin_client, admin_user, user_factory, sort):
    user_factory.create_batch(5)

    resp = admin_client.get(reverse("user-list"), {"sort": sort})

    expected = list(
        User.objects.all().order_by(sort).values_list("username", flat=True)
    )

    assert expected == [d["attributes"]["username"] for d in resp.json()["data"]]



from django.db.models.functions import Lower
from django_filters import FilterSet
from django_filters.filters import CharFilter
from rest_framework import filters

from emeis.core.models import User


class UserFilterset(FilterSet):
    has_role = CharFilter(
        field_name="acls__role__pk",
        distinct=True,
    )

    class Meta:
        model = User

        fields = {
            "id": ["exact", "in"],
            "username": ["exact", "in"],
            "first_name": ["exact", "icontains", "contains"],
            "last_name": ["exact", "icontains", "contains"],
            "email": ["exact", "in"],
        }


class CaseInsensitiveOrderingFilter(filters.OrderingFilter):
    def _make_ordering_field(self, field, view):
        desc = field.startswith("-")
        field_name = field[1:] if desc else field

        if field_name in getattr(view, "case_insensitive_ordering_fields", []):
            return Lower(field_name).desc() if desc else Lower(field_name)
        return field

    def get_ordering(self, request, queryset, view):
        ordering = super().get_ordering(request, queryset, view)
        if not ordering:
            return ordering
        return [self._make_ordering_field(field, view) for field in ordering]



# Generated by Django 2.2.13 on 2020-06-12 08:15
from django.db import migrations


class Migration(migrations.Migration):
    # This migration used to do things, but doesn't anymore.
    # We cannot remove it due to keeping migration state consistent

    dependencies = [
        ("emeis_core", "0001_initial"),
    ]

    operations = [
        # migrations.RunPython(migrations.RunPython.noop, migrations.RunPython.noop),
    ]



import pytest
from django.apps import apps
from django.urls import reverse

from emeis.oidc_auth.authentication import OIDCUser


@pytest.mark.parametrize("requesting_user", ["anon", "user", "admin"])
def test_own_and_admin_visibility(
    db,
    user_factory,
    scope_factory,
    role_factory,
    permission_factory,
    acl_factory,
    admin_user,
    settings,
    client,
    user,
    requesting_user,
):

    settings.GENERIC_PERMISSIONS_VISIBILITY_CLASSES = [
        "emeis.core.visibilities.OwnAndAdmin"
    ]
    apps.get_app_config("generic_permissions").ready()

    admin_role = role_factory(slug="admin")

    other_user = user_factory()

    scope1 = scope_factory()
    role1 = role_factory()
    perm1 = permission_factory()
    permission_factory()
    role1.permissions.add(perm1)
    acl_factory(user=user, scope=scope1, role=role1)

    expected_count = 0
    if requesting_user == "admin":
        client.force_authenticate(
            OIDCUser(username=admin_user.username, claims={"sub": admin_user.username})
        )
        user.acls.create(role=admin_role, scope=scope_factory())
        expected_count = 2
    elif requesting_user == "user":
        client.force_authenticate(
            OIDCUser(username=other_user.username, claims={"sub": other_user.username})
        )
        acl_factory(user=other_user, role=role1)
        expected_count = 1

    resp = client.get(reverse("user-list"))
    assert len(resp.json()["data"]) == expected_count

    resp = client.get(reverse("scope-list"))
    assert len(resp.json()["data"]) == expected_count

    resp = client.get(reverse("role-list"))
    assert len(resp.json()["data"]) == expected_count

    resp = client.get(reverse("permission-list"))
    assert len(resp.json()["data"]) == expected_count

    resp = client.get(reverse("acl-list"))
    assert len(resp.json()["data"]) == (2 if requesting_user == "admin" else 0)
